# -*- coding: utf-8 -*-
"""
export2kml.py
QGIS plugin: Export 2KML

Batch‐export selected vector and raster layers to a single KML or KMZ.
Vectors become <Placemark> entries; rasters are rendered with their QML style,
saved as opaque PNGs, and embedded as <GroundOverlay> in the KMZ.
"""

import os
import tempfile
import zipfile
import xml.etree.ElementTree as ET

from qgis.PyQt import uic
from qgis.PyQt.QtWidgets import (
    QAction,
    QFileDialog,
    QTableWidgetItem,
    QCheckBox,
    QComboBox,
    QMessageBox,
    QToolBar
)
from qgis.PyQt.QtGui import QIcon, QImage, QPainter
from qgis.PyQt.QtCore import Qt, QSize
from qgis.core import (
    QgsProject,
    QgsVectorLayer,
    QgsRasterLayer,
    QgsMapSettings,
    QgsMapRendererParallelJob,
    Qgis
)
from osgeo import ogr, gdal, osr

gdal.UseExceptions()
ogr.UseExceptions()

KML_PROPS = [
    'Name', 'description', 'timestamp', 'begin', 'end',
    'altitudeMode', 'tessellate', 'extrude', 'visibility',
    'drawOrder', 'icon'
]


def make_kml_root():
    kml = ET.Element('kml', xmlns='http://www.opengis.net/kml/2.2')
    doc = ET.SubElement(kml, 'Document')
    return kml, doc


def add_vector_layer(doc, path, props):
    ds = ogr.Open(path)
    if ds is None:
        raise RuntimeError(f'Cannot open vector {path}')
    layer = ds.GetLayer(0)

    folder_label = props.get('folder_name') or os.path.splitext(os.path.basename(path))[0]
    folder = ET.SubElement(doc, 'Folder')
    ET.SubElement(folder, 'name').text = folder_label

    ld = layer.GetLayerDefn()
    avail = [ld.GetFieldDefn(i).GetName() for i in range(ld.GetFieldCount())]

    for feat in layer:
        pm = ET.SubElement(folder, 'Placemark')
        if props.get('name'):
            nm = props['name']
            val = feat.GetField(nm) if nm in avail else nm
            ET.SubElement(pm, 'name').text = str(val)
        if props.get('description'):
            d = props['description']
            val = feat.GetField(d) if d in avail else d
            ET.SubElement(pm, 'description').text = str(val)
        if props.get('fields'):
            ext = ET.SubElement(pm, 'ExtendedData')
            for fld in props['fields']:
                if fld in avail:
                    dt = ET.SubElement(ext, 'Data', name=fld)
                    ET.SubElement(dt, 'value').text = str(feat.GetField(fld))
        geom = feat.GetGeometryRef()
        if geom:
            k = ogr.CreateGeometryFromWkb(geom.ExportToWkb())
            wrap = ET.fromstring(f'<r>{k.ExportToKML()}</r>')
            for c in wrap:
                pm.append(c)
    ds = None


def add_raster_layer(doc, path, props, rasters_dir):
    os.makedirs(rasters_dir, exist_ok=True)
    base = os.path.splitext(os.path.basename(path))[0]
    img_name = f'{base}.png'
    img_path = os.path.join(rasters_dir, img_name)

    rlayer = QgsRasterLayer(path, base)
    if not rlayer.isValid():
        raise RuntimeError(f'Cannot load raster {path}')

    extent = rlayer.extent()
    width, height = rlayer.width(), rlayer.height()
    ms = QgsMapSettings()
    ms.setLayers([rlayer])
    ms.setExtent(extent)
    ms.setOutputSize(QSize(width, height))

    job = QgsMapRendererParallelJob(ms)
    job.start()
    job.waitForFinished()

    img: QImage = job.renderedImage()
    rgb = QImage(width, height, QImage.Format_RGB32)
    rgb.fill(Qt.white)
    painter = QPainter(rgb)
    painter.drawImage(0, 0, img)
    painter.end()
    rgb.save(img_path, 'PNG')

    ds = gdal.Open(path)
    if ds is None:
        raise RuntimeError(f'Cannot open raster {path}')
    src = osr.SpatialReference(wkt=ds.GetProjection())
    tgt = osr.SpatialReference(); tgt.ImportFromEPSG(4326)
    if not src.IsSame(tgt):
        ds = gdal.Warp('', ds, format='VRT', dstSRS='EPSG:4326')
    gt = ds.GetGeoTransform()
    xs, ys = ds.RasterXSize, ds.RasterYSize
    minx, maxy = gt[0], gt[3]
    maxx = minx + gt[1] * xs
    miny = maxy + gt[5] * ys
    ds = None

    go = ET.SubElement(doc, 'GroundOverlay')
    ET.SubElement(go, 'visibility').text = '1'
    label = props.get('folder_name') or props.get('name') or base
    ET.SubElement(go, 'name').text = label
    icon = ET.SubElement(go, 'Icon')
    ET.SubElement(icon, 'href').text = f'rasters/{img_name}'
    ll = ET.SubElement(go, 'LatLonBox')
    ET.SubElement(ll, 'north').text = str(maxy)
    ET.SubElement(ll, 'south').text = str(miny)
    ET.SubElement(ll, 'east').text = str(maxx)
    ET.SubElement(ll, 'west').text = str(minx)

    return img_path


class Export2KML:
    def __init__(self, iface):
        self.iface = iface
        self.plugin_dir = os.path.dirname(__file__)
        self.added_file_layers = []
        # Menu and toolbar identifiers
        self.menu = 'MAS Vector Processing'
        self.toolbar = None
        self.actions = []

    def initGui(self):
        ui_path = os.path.join(self.plugin_dir, 'form.ui')
        self.dlg = uic.loadUi(ui_path)

        # Find or create common toolbar
        self.toolbar = self.iface.mainWindow().findChild(QToolBar, 'MASVectorProcessingToolbar')
        if self.toolbar is None:
            self.toolbar = self.iface.addToolBar('MAS Vector Processing')
            self.toolbar.setObjectName('MASVectorProcessingToolbar')

        # Create action
        icon = QIcon(os.path.join(self.plugin_dir, 'export2kml.png'))
        self.action = QAction(icon, 'Export 2KML…', self.iface.mainWindow())
        self.action.triggered.connect(self.show_dialog)

        # Add to common Vector menu and toolbar
        self.iface.addPluginToVectorMenu(self.menu, self.action)
        self.toolbar.addAction(self.action)
        self.actions.append(self.action)

        # Connect UI signals
        self.dlg.btnBrowseFiles.clicked.connect(self.add_layer)
        self.dlg.btnSelectOutput.clicked.connect(self.select_output)
        self.dlg.btnRun.clicked.connect(self.run_export)
        self.dlg.btnCancel.clicked.connect(self.dlg.close)

    def unload(self):
        for action in self.actions:
            self.iface.removePluginMenu(self.menu, action)
            self.iface.removeToolBarIcon(action)
        self.actions.clear()

    def show_dialog(self):
        self.populate_table()
        self.update_format_options()
        self.dlg.setParent(self.iface.mainWindow(), Qt.Tool)
        self.dlg.setWindowFlags(self.dlg.windowFlags() | Qt.Tool)
        self.dlg.show()

    def populate_table(self):
        project_layers = list(QgsProject.instance().mapLayers().values())
        all_layers = project_layers + self.added_file_layers
        tbl = self.dlg.tableLayers
        tbl.setColumnCount(3 + len(KML_PROPS))
        tbl.setHorizontalHeaderLabels(['Export?', 'Layer Name', 'Type'] + KML_PROPS)
        tbl.setRowCount(len(all_layers))

        for i, lyr in enumerate(all_layers):
            cb = QCheckBox()
            cb.setChecked(True)
            cb.stateChanged.connect(self.update_format_options)
            tbl.setCellWidget(i, 0, cb)

            item_name = QTableWidgetItem(lyr.name())
            item_name.setFlags(item_name.flags() & ~Qt.ItemIsEditable)
            tbl.setItem(i, 1, item_name)

            typ = 'Raster' if isinstance(lyr, QgsRasterLayer) else 'Vector'
            item_type = QTableWidgetItem(typ)
            item_type.setFlags(item_type.flags() & ~Qt.ItemIsEditable)
            tbl.setItem(i, 2, item_type)

            for j in range(len(KML_PROPS)):
                combo = QComboBox()
                combo.addItem('<None>')
                if isinstance(lyr, QgsVectorLayer):
                    combo.addItems([f.name() for f in lyr.fields()])
                combo.setEditable(True)
                tbl.setCellWidget(i, j + 3, combo)

    def update_format_options(self):
        tbl = self.dlg.tableLayers
        any_raster = False
        project_layers = list(QgsProject.instance().mapLayers().values())
        all_layers = project_layers + self.added_file_layers
        for i, lyr in enumerate(all_layers):
            cb = tbl.cellWidget(i, 0)
            if isinstance(lyr, QgsRasterLayer) and cb.isChecked():
                any_raster = True
                break
        self.dlg.radioKML.setEnabled(not any_raster)
        if any_raster:
            self.dlg.radioKMZ.setChecked(True)

    def add_layer(self):
        path, _ = QFileDialog.getOpenFileName(
            self.dlg, 'Browse Input Files', '',
            'Vector (*.geojson *.shp *.gpkg);;Raster (*.jpg *.png *.tif *.tiff);;All (*.*)'
        )
        if not path:
            return
        if path.lower().endswith(('.jpg', '.png', '.tif', '.tiff')):
            lyr = QgsRasterLayer(path, os.path.basename(path))
        else:
            lyr = QgsVectorLayer(path, os.path.basename(path), 'ogr')
        if not lyr.isValid():
            QMessageBox.critical(self.iface.mainWindow(), 'Error', 'Cannot load layer')
            return
        self.added_file_layers.append(lyr)
        self.populate_table()
        self.update_format_options()

    def select_output(self):
        fmt = 'kmz' if self.dlg.radioKMZ.isChecked() else 'kml'
        path, _ = QFileDialog.getSaveFileName(self.dlg, 'Output File', '', f'*.{fmt}')
        if not path:
            return
        if not path.lower().endswith(f'.{fmt}'):
            path += f'.{fmt}'
        self.output_path = path
        self.dlg.editOutput.setText(path)

    def run_export(self):
        try:
            out = getattr(self, 'output_path', None)
            if not out:
                raise ValueError('Please select an output file')

            kml, doc = make_kml_root()
            tmp = tempfile.mkdtemp()
            rasters_dir = os.path.join(tmp, 'rasters')
            os.makedirs(rasters_dir, exist_ok=True)
            collected = []

            progress = getattr(self.dlg, 'progressBar', None)
            if progress:
                progress.setValue(0)
                progress.setVisible(True)

            project_layers = list(QgsProject.instance().mapLayers().values())
            all_layers = project_layers + self.added_file_layers
            tbl = self.dlg.tableLayers

            total_rasters = sum(
                1 for i, lyr in enumerate(all_layers)
                if isinstance(lyr, QgsRasterLayer) and tbl.cellWidget(i, 0).isChecked()
            )
            raster_count = 0

            for i, lyr in enumerate(all_layers):
                if not tbl.cellWidget(i, 0).isChecked():
                    continue
                path = lyr.source()
                props = {}
                for j, prop in enumerate(KML_PROPS, start=3):
                    text = tbl.cellWidget(i, j).currentText().strip()
                    if text and text != '<None>':
                        props[prop.lower()] = text

                if isinstance(lyr, QgsRasterLayer):
                    png_path = add_raster_layer(doc, path, props, rasters_dir)
                    collected.append(png_path)
                    raster_count += 1
                    if progress and total_rasters:
                        pct = int(raster_count / total_rasters * 100)
                        progress.setValue(pct)
                else:
                    add_vector_layer(doc, path, props)

            kml_file = os.path.join(tmp, 'doc.kml')
            ET.ElementTree(kml).write(kml_file, encoding='utf-8', xml_declaration=True)

            if collected:
                with zipfile.ZipFile(out, 'w', zipfile.ZIP_DEFLATED) as zf:
                    zf.write(kml_file, 'doc.kml')
                    for f in collected:
                        zf.write(f, f'rasters/{os.path.basename(f)}')
            else:
                os.replace(kml_file, out)

            if progress:
                progress.setVisible(False)

            self.iface.messageBar().pushMessage(
                'Export 2KML',
                f'Exported to {out}',
                level=Qgis.Info
            )
            self.dlg.close()

        except Exception as e:
            QMessageBox.critical(self.iface.mainWindow(), 'Error', str(e))
